// ;-*-C++-*- *  Time-stamp: "2008-07-01 17:34:10 hmmr"

/***************************************************************************
 *            agh_data.cc
 *
 *  Copyright  2008  Andrei Zavada
 *  Email johnhommer@gmail.com

 *  Portions of code taken from the wfdb library, with the following
 *  copyright notice:
 *  -----------------
 * file: fft.c		G. Moody	24 February 1988
		   Last revised:	  10 June 2005

-------------------------------------------------------------------------------
fft: Fast Fourier transform of real data
Copyright (C) 1988-2005 George B. Moody

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA.

You may contact the author by e-mail (george@mit.edu) or postal mail
(MIT Room E25-505A, Cambridge, MA 02139 USA).  For updates to this software,
please visit PhysioNet (http://www.physionet.org/).

 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <math.h>
#include "agh_expdesign.hh"
#include "agh_io.hh"
#include "agh_data.hh"


#define PI	M_PI	/* pi to machine precision, defined in math.h */
#define TWOPI	(2.0*PI)



static void detrend( double*, int);

static void four1( double *data, int nn);
static void realft( double *data, int n);




/* See Oppenheim & Schafer, Digital Signal Processing, p. 241 (1st ed.) */
inline double win_bartlett( int j, int n, double *wsum)
{
	double a = 2.0/(n-1), w;
	if ( (w = j*a) > 1.0 )
		w = 2.0 - w;
	*wsum += w;
	return w;
}

/* See Oppenheim & Schafer, Digital Signal Processing, p. 242 (1st ed.) */
inline double win_blackman( int j, int n, double *wsum)
{
	double a = TWOPI/(n-1), w;
	w = 0.42 - 0.5*cos(a*j) + 0.08*cos(2*a*j);
	*wsum += w;
	return w;
}

/* See Harris, F.J., "On the use of windows for harmonic analysis with the
   discrete Fourier transform", Proc. IEEE, Jan. 1978 */
inline double win_blackman_harris( int j, int n, double *wsum)
{
	double a = TWOPI/(n-1), w;
	w = 0.35875 - 0.48829*cos(a*j) + 0.14128*cos(2*a*j) - 0.01168*cos(3*a*j);
	*wsum += w;
	return w;
}

/* See Oppenheim & Schafer, Digital Signal Processing, p. 242 (1st ed.) */
inline double win_hamming( int j, int n, double *wsum)
{
	double a = TWOPI/(n-1), w;
	w = 0.54 - 0.46*cos(a*j);
	*wsum += w;
	return w;
}

/* See Oppenheim & Schafer, Digital Signal Processing, p. 242 (1st ed.)
   The second edition of Numerical Recipes calls this the "Hann" window. */
inline double win_hanning( int j, int n, double *wsum)
{
	double a = TWOPI/(n-1), w;
	w = 0.5 - 0.5*cos(a*j);
	*wsum += w;
	return w;
}

/* See Press, Flannery, Teukolsky, & Vetterling, Numerical Recipes in C,
   p. 442 (1st ed.) */
inline double win_parzen( int j, int n, double *wsum)
{
	double a = (n-1)/2.0, w;
	if ( (w = (j-a)/(a+1)) > 0.0 )
		w = 1 - w;
	else
		w = 1 + w;
	*wsum += w;
	return w;
}

/* See any of the above references. */
inline double win_square( int j, int n, double *wsum)
{
	*wsum += 1.0;
	return 1.0;
}

/* See Press, Flannery, Teukolsky, & Vetterling, Numerical Recipes in C,
   p. 442 (1st ed.) or p. 554 (2nd ed.) */
inline double win_welch( int j, int n, double *wsum)
{
	double a = (n-1)/2.0, w;
	w = (j-a)/(a+1);
	w = 1 - w*w;
	*wsum += w;
	return w;
}


// static (window*)( double, guint, guint, double*)[8] = {
//	win_bartlett,
//	win_blackman,
//	win_blackman_harris,
//	win_hamming,
//	win_hanning,
//	win_parzen,
//	win_square,
//	win_welch
// };





static void detrend( double *data, guint n)
{
	int	i;
	double	a, b = 0.0,
		tsqsum = 0.0, ysum = 0.0, t;

	for ( i = 0; i < n; i++ )
		ysum += data[i];
	for ( i = 0; i < n; i++ ) {
		t = i - n/2 + 0.5;
		tsqsum += t*t;
		b += t * data[i];
	}
	b /= tsqsum;
	a = ysum/n - b*(n-1)/2.0;
	for ( i = 0; i < n; i++ )
		data[i] -= a + b*i;
	if ( b < -0.04 || b > 0.04 )
		printf( "possibly significant trend in input series\n");
}














inline static void realft( double *data, int n)
{
	int	i, i1, i2, i3, i4,
		n2p3;
	double	c1 = 0.5, c2 = -0.5,
		h1r, h1i, h2r, h2i;
	double	wr, wi, wpr, wpi, wtemp,
		theta = PI/(double)n;

	four1( data, n);

	wtemp = sin(0.5*theta);
	wpr = -2.0 * wtemp*wtemp;
	wpi = sin(theta);
	wr = 1.0+wpr;
	wi = wpi;
	n2p3 = 2*n+3;
	for (i = 2; i <= n/2; i++) {
		i4 = 1 + (i3 = n2p3 - (i2 = 1 + ( i1 = i + i - 1)));
		h1r =  c1*(data[i1] + data[i3]);
		h1i =  c1*(data[i2] - data[i4]);
		h2r = -c2*(data[i2] + data[i4]);
		h2i =  c2*(data[i1] - data[i3]);
		data[i1] =  h1r + wr*h2r - wi*h2i;
		data[i2] =  h1i + wr*h2i + wi*h2r;
		data[i3] =  h1r - wr*h2r + wi*h2i;
		data[i4] = -h1i + wr*h2i + wi*h2r;
		wr = (wtemp = wr)*wpr - wi*wpi+wr;
		wi = wi*wpr + wtemp*wpi + wi;
	}

	data[1] = (h1r = data[1]) + data[2];
	data[2] = h1r - data[2];
}





static void four1( double *data, int nn)
{
	int n, mmax, m, j, istep, i;
	double wtemp, wr, wpr, wpi, wi, theta;
	float tempr, tempi;

	n = nn << 1;
	j = 1;
	for (i = 1; i < n; i += 2) {
		if (j > i) {
			tempr = data[j];     data[j] = data[i];     data[i] = tempr;
			tempr = data[j+1]; data[j+1] = data[i+1]; data[i+1] = tempr;
		}
		m = n >> 1;
		while (m >= 2 && j > m) {
			j -= m;
			m >>= 1;
		}
		j += m;
	}
	mmax = 2;
	while ( n > mmax ) {
		istep = 2*mmax;
		theta = TWOPI/mmax;
		wtemp = sin(0.5*theta);
		wpr = -2.0*wtemp*wtemp;
		wpi = sin(theta);
		wr = 1.0;
		wi = 0.0;
		for ( m = 1; m < mmax; m += 2 ) {
			for ( i = m; i <= n; i += istep ) {
				j = i + mmax;
				tempr = wr*data[j]   - wi*data[j+1];
				tempi = wr*data[j+1] + wi*data[j];
				data[j]   = data[i]   - tempr;
				data[j+1] = data[i+1] - tempi;
				data[i] += tempr;
				data[i+1] += tempi;
			}
			wr = (wtemp = wr)*wpr - wi*wpi + wr;
			wi = wi*wpr + wtemp*wpi + wi;
		}
		mmax = istep;
	}
}







inline static int
__max_power_of_2_less_than_or_equal( gint x)
{
	gint t = 1;
	do  t <<= 1;  while ( t < x );
	return t;
}




gint
CBinnedPower::obtain_power( CEDFFile *F, gint sig_no,
			    gint ipagesize, SFFTParamsSet& req_params,
			    gboolean redo_filtered)
{
	guint	window_size = 2 << req_params.window_size_2power;
	if ( ipagesize < window_size ) {
		printf( "CBinnedPower::obtain_power: window_size > page_size (%d > %d)\n",
			window_size, ipagesize);
		return -5;
	}

	gfloat	bin_size_effective = 1./window_size; // samplerate / smpl_per_window;
	gint	lump_factor = req_params.bin_size / bin_size_effective;
	if ( lump_factor * bin_size_effective != req_params.bin_size ) {
		printf( "CBinnedPower::obtain_power: requested bin_size (%g) not a multiple of effective bin_size (%g)\n",
			req_params.bin_size, bin_size_effective);
		return -5;
	}

	_samplerate = F->H(sig_no)->SamplesPerRecord / F->DataRecordSize;
	_pagesize   = ipagesize;
	fft_params  = req_params;

//	fft_params.dump();

	guint	smpl_per_page = _pagesize * _samplerate,
		smpl_per_window = __max_power_of_2_less_than_or_equal( window_size * _samplerate);
//		smpl_per_window = window_size * _samplerate;

	GArray	*S = g_array_new( FALSE, FALSE, sizeof(double));
	if ( F->ship_data( S, sig_no, 0, F->DataRecordSize * F->NDataRecords) )
		return -1;

	gulong	pages = F->DataRecordSize * F->NDataRecords / _pagesize,
		total_samples = pages * _samplerate * _pagesize,
		p = 0, w_pos_smpl = 0;
	guint	page_window_cnt = 0,
		n_bins = __max_power_of_2_less_than_or_equal( _samplerate/2/fft_params.bin_size);

	printf( "samplerate: %d  lump_factor: %d\n"
		"pages: %ld  S->len = %d  n_bins = %d  total_samples = %d  smpl_per_window = %d\n",
		_samplerate, lump_factor,
		pages, S->len, n_bins, total_samples, smpl_per_window);

	double	*page_pow_accum = (double*) g_malloc( sizeof(double) * (n_bins + 1));  // +1 added to fix the case of samplerate not being an exact power of 2

	set_size( pages+2);

	if ( status & AGH_BP_HAS_TRACK ) {
		gulong	len_sec = _pages * _pagesize;
		for ( gulong s = 0; s < len_sec; s++ )
			if ( nth_second(s) == 1 )
				for ( guint smpl = 0; smpl < _samplerate; smpl++ )
					g_array_index (S, double, (s * _samplerate) + smpl) /= 50;
	} else
		memset( _track, 0, _pages * _pagesize * sizeof(gchar));

	double	*vec = (double*) g_malloc( sizeof(double) * (smpl_per_window + 2));  // +1
	double	rsum = 0., rmean = 0.,
		wsum = 0.,
		norm;

	p = 0;
	while ( w_pos_smpl < total_samples ) {
		memcpy( vec, &g_array_index (S, double, w_pos_smpl),
			     (smpl_per_window+2) * sizeof(double));
		guint i;
	      // zero-mean and detrend
		for ( i = 0, rsum = 0.; i < smpl_per_window; i++ )
			rsum += vec[i];
		rmean = rsum / smpl_per_window;
		for ( i = 0; i < smpl_per_window; i++)
			vec[i] -= rmean;
		detrend( vec, smpl_per_window);

	      // apply windowing function
		switch ( fft_params.window_type ) {
		case AGH_WT_BARTLETT:
			for ( i = 0, wsum = 0.; i < smpl_per_window; i++ )
				vec[i] *= win_bartlett( i, smpl_per_window, &wsum);
		    break;
		case AGH_WT_BLACKMAN:
			for ( i = 0, wsum = 0.; i < smpl_per_window; i++ )
				vec[i] *= win_blackman( i, smpl_per_window, &wsum);
		    break;
		case AGH_WT_BLACKMAN_HARRIS:
			for ( i = 0, wsum = 0.; i < smpl_per_window; i++ )
				vec[i] *= win_blackman_harris( i, smpl_per_window, &wsum);
		    break;
		case AGH_WT_HAMMING:
			for ( i = 0, wsum = 0.; i < smpl_per_window; i++ )
				vec[i] *= win_hamming( i, smpl_per_window, &wsum);
		    break;
		case AGH_WT_HANNING:
			for ( i = 0, wsum = 0.; i < smpl_per_window; i++ )
				vec[i] *= win_hanning( i, smpl_per_window, &wsum);
		    break;
		case AGH_WT_PARZEN:
			for ( i = 0, wsum = 0.; i < smpl_per_window; i++ )
				vec[i] *= win_parzen( i, smpl_per_window, &wsum);
		    break;
		case AGH_WT_SQUARE:
			for ( i = 0, wsum = 0.; i < smpl_per_window; i++ )
				vec[i] *= win_square( i, smpl_per_window, &wsum);
		    break;
		case AGH_WT_WELCH:
			for ( i = 0, wsum = 0.; i < smpl_per_window; i++ )
				vec[i] *= win_welch( i, smpl_per_window, &wsum);
		    break;
		}
		norm = sqrt( 2. / (wsum * smpl_per_window));

	      // here goes magic
	      // weird thing here is that we pass a pointer to outside the buffer
		realft( vec-1, smpl_per_window/2);	/* perform the FFT;  see Numerical Recipes */

	      // collect and accumulate power
		vec[smpl_per_window] = vec[1];		/* unpack the output array */
		vec[1] = vec[smpl_per_window+1] = 0.;   // not a slighest idea what this is for
		guint	bin, bin_incr = 2 * lump_factor;
		for ( i = 0, bin = 0; i < smpl_per_window; i += bin_incr, bin++ )
			for ( guint j = 0; j < 2*lump_factor; j += 2)
				page_pow_accum[bin] += (pow( vec[i+j], 2) + pow( vec[i+j+1], 2)) * norm*norm;


		w_pos_smpl += smpl_per_window/2;
		page_window_cnt++;

	      // average and save on new page
		if ( w_pos_smpl > p * smpl_per_page ) {
			for ( guint bin = 0; bin < n_bins; bin++ ) {
				ijth_bin( bin, p) = page_pow_accum[bin] / page_window_cnt;
//					if ( p < 20 && bin < 10 )
//						printf( "bin[%d,%d] = %g\n", bin, p, ijth_bin( bin, p));
				page_pow_accum[bin] = 0.;
			}
			page_window_cnt = 0;
			p++;
		}
	}

	g_free( page_pow_accum);
	g_free( vec);
	g_array_free( S, TRUE);

	status |= (AGH_BP_OBTAINED | AGH_BP_INMEM);

	return 0;
}

// EOF
