// ;-*-C++-*- *  Time-stamp: "2008-06-21 15:06:03 hmmr"

/***************************************************************************
 *            z_lists.hh
 *
 *  Copyright  2008  Andrei Zavada
 *  Email johnhommer@gmail.com
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef _Z_LISTS_H
#define _Z_LISTS_H


#if HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glib.h>
#include <stdio.h>


// this is used to tell ctors and dtors of objects registrable in
// CLists to create a lite version of an object, just sufficient, when
// compared to an existing object using operator ==, to obtain a
// pointer to the valid copy
#define PHONY TRUE

// and this is a flag in an object's status field to indicate such
// condition
#define Z_OBJECT_PHONY (1 << 8)



class COrderedName;

template<class> class CList;
template<class> class CRefList;
template<class> class CAggregatorList;





// a generial-purpose class for named objects to appear in lists in an
// ordered fashion

class COrderedName {
    friend class CList<COrderedName>;

	GQuark _q;
	gint _seq_no;
    public:
	const gchar *name()  { return g_quark_to_string( _q); }
	guint seq_no()       { return _seq_no; }

	COrderedName()  { _q = 0, _seq_no = -1; }
	COrderedName( const gchar *n,
		      gint seq_no = -1)  { _q = g_quark_from_string( n), _seq_no = seq_no; }
	COrderedName( GQuark q,
		      gint seq_no = -1)  { _q = q, _seq_no = seq_no; }

	void rename( const gchar *new_name)  { _q = g_quark_from_string( new_name); }

	void dump()  { printf( "%s\t: %d\n", g_quark_to_string( _q), _seq_no); }

      // only consider _q
	gboolean operator == ( COrderedName &o)  { return _q == o._q; }
};








template<class O> class CRefList {
    friend class CAggregatorList<O>;

    private:
	GString *_name;
	GString *_comment;
	GPtrArray *_list;

    public:
	CRefList( const gchar *iname,
		  const gchar *icomment = NULL)	 { _name = g_string_new( iname);
						   _comment = g_string_new( icomment);
						   _list = g_ptr_array_sized_new( 20);	}

	~CRefList()  {	g_string_free( _name, TRUE);
			g_string_free( _comment, TRUE);
			for ( guint i = 0; i < _list->len; i++ )
				nth(i)->belongs = NULL;
			g_ptr_array_free( _list, TRUE);	          }

	gchar *name()     { return _name->str; };
	gchar *comment()  { return _comment->str; }

	guint n()         { return _list->len; }
	O *nth( guint i)  { return (O*) g_ptr_array_index (_list, i); }

	void rename( const gchar *name)  { g_string_assign( _name, name); }

	void checkin ( O *j)  { g_ptr_array_remove( _list, (gpointer)j); g_ptr_array_add( _list, (gpointer)j); }
	void checkout( O *j)  { g_ptr_array_remove( _list, (gpointer)j); }

      // for this to be useful, we require default ctors for O that do not have side-effects:
      // what an inconvenience!
	O *exist( O *what)	{ for ( guint i = 0; i < n(); i++ )
				   { O *o = nth(i);
				     if ( *o == *what ) // and of course, the == operator in every class instance
					return o;        }
				  return NULL;                       }
	void dump()
	{
		printf( "CRefList \"%s\" referencing %d elements:\n", _name->str, n());
		for ( guint i = 0; i < n(); i++ ) {
			printf( "  [%d]", i);
			nth(i)->dump();
		}
	}
};








// ----------------------------------------------------------------------------
// Now, CList acting as a db table. Objects are stored in a linked
// list not ptr_array, such that their locations (referenced from
// CList instantiations) remain constant

template<class O> class CList {
// friend

	GSList *list;
	guint _n;

    public:
	CList()	{ list = NULL; _n = 0; }
       ~CList()	{ GSList *node = list;
		  while ( node )
		    { delete (O*) (node->data);
		      node = node->next;       }
		  g_slist_free( list);           }

	guint n()		{ return _n; }
	O *nth( guint i)	{ return (O*) g_slist_nth_data( list, i); }

	O *exist( O *o)		{ GSList *node = list;
				  while ( node )
				      if ( *(O*) (node->data) == *o )	return (O*) (node->data);
				      else				node = node->next;
				  return NULL; }

	gint add( O *o)		{ if ( !exist( o) )  { list = g_slist_append( list, (gpointer)o); _n++; return 1; }
				  else                 return 0; }
	gint ins( O *o, guint at);

	gint del( O *o)		{ GSList *node = list;
				  while ( node )
				      if ( *(O*) (node->data) == *o )  { delete (O*) (node->data);
									 list = g_slist_delete_link( list, node);
									 _n--;
									 return 1; }
				      else			       node = node->next;
				  return 0; }

	void del_nth( guint i)  { GSList *node = g_slist_nth( list, i);
				  delete (O*) (node->data);
				  list = g_slist_delete_link( list, node); }

	void del_all()          { GSList *node = list;
				  while ( node )
				      { delete (O*) (node->data);
					node = node->next;        }
				  g_slist_free( list);
				  list = NULL, _n = 0;              }

	void clean_unowned()
	{
		GSList *node = list, *nextnode;
		while ( node ) {
			O *o = (O*) (node->data);
			if ( o -> belongs == NULL ) {
				delete o;
				nextnode = node->next;
				list = g_slist_delete_link( list, node);
				node = nextnode;
				_n--;
			} else
				node = node->next;
		}
	}

	void dump()
	{
		printf( "A CList holding %d elements:\n", n());
		for ( guint i = 0; i < n(); i++ ) {
			printf( "  <%d>\t", i);
			nth(i)->dump();
		}
	}

      // various odd member function specializations
	COrderedName *o_by_seq_no( gint);
	COrderedName *o_by_name( const gchar *);
	gint swap( gint, gint);
};



// we need to assign a proper _seq_no, which for this case will be the
// n() of the containing list; and, o._seq_no is ignored
template<> inline gint
CList<COrderedName>::add( COrderedName *o)
{
	if ( !exist( o) ) {
		list = g_slist_append( list, (gpointer)o);
		o->_seq_no = _n++; // adding this line makes all the diff
		return 1;
	} else
		return 0;
}


template<> inline gint
CList<COrderedName>::ins( COrderedName *o, guint at)
{
	if ( !exist( o) ) {
		o->_seq_no = at;
		list = g_slist_append( list, (gpointer)o);

		GSList *node = list;
		while ( node && node->next ) { // thus stopping right before the newly appended one
			if ( ((COrderedName*)(node->data))->_seq_no >= at )
				((COrderedName*)(node->data))->_seq_no++;
			node = node->next;
		}
		_n++;
		return 1;
	} else
		return 0;
}







template<class L> class CAggregatorList {
	GPtrArray *array;

    public:
	CAggregatorList()  { array = g_ptr_array_sized_new( 3); }
       ~CAggregatorList()  { for ( guint i = 0; i < array->len; i++ )
				 delete (L*) g_ptr_array_index (array, i);
			     g_ptr_array_free( array, TRUE); }

	guint n()	   { return array->len; }
	L *nth( guint i)   { return (L*) g_ptr_array_index (array, i); }

	L *o_by_name( const gchar *name)	{ for ( guint i = 0; i < n(); i++ ) {
						      L *l = nth(i);
						      if ( !strcmp( l->name(), name) )
							      return l;
						  }
						  return NULL; }

	gint add( const gchar *name,
		  const gchar *comment = NULL)	{ if ( o_by_name( name) )                                    return 0;
						  g_ptr_array_add( array, (gpointer) new L( name, comment)); return 1; }
	void del( const gchar *name)	{ for ( guint i = 0; i < n(); i++ )
					      if ( !strcmp( nth(i)->name(), name) ) {
						      delete (L*) g_ptr_array_remove_index (array, i);
						      return;
					      }  }
	void del_nth( guint i)		{ delete (L*) g_ptr_array_remove_index (array, i); }

	void dump()
	{
		printf( "A CAggregatorList holding %d elements:\n", n());
		for ( guint i = 0; i < n(); i++ ) {
			printf( "  <%d>\t", i);
			nth(i)->dump();
		}
	}
};






//class CLinkedObject {
//
//  public:
//	CRefList<CLinkedObject> *belongs;
//	int status;
//
//	CLinkedObject()                              {  status = Z_OBJECT_PHONY, belongs = NULL;    }
//
//	CLinkedObject( CRefList<CLinkedObject> *to)  {  status = 0, belongs = NULL;                 }
//			(belongs = to)->checkin ( this);
//			printf( "\t\tcheckin %p\n", this); } }
//
//       ~CLinkedObject()                              {  if ( belongs && status | ~Z_OBJECT_PHONY )
//								   belongs      ->checkout( this);  }
//
//	virtual bool operator == ( CLinkedObject &o);   // not used for CLinkedObject's themselves
//};



#endif

// EOF
